import os
import time
import logging

from pathlib import Path
from logging import (FileHandler,
                     StreamHandler,
                     Formatter,
                     Logger)

from source.testcase import Status


date_format = '%Y-%m-%d %H:%M:%S'
text_format = '%(asctime)s %(levelname)-8s ' \
              'line:%(lineno)-2s of %(filename)-15s >> %(message)s'


class S_FileHandler(FileHandler):
    """
    Class that overrides standard Logging FileHanlder.
    """

    def __init__(self,
                 test_name: str,
                 session_date: str,
                 mode: str = 'a',
                 encoding: str | None = None,
                 delay: int = 0
                 ) -> None:
        """
        Initialization of FileHandler instance.

        Args:
            test_name (str): From test module, starts with "test_*".
            session_date (str): Start date of whole test session.
            mode (str, optional): File read/write/append. Defaults to 'a'.
            encoding (str | None, optional): Type of file encoding.
            delay (int, optional): If true, file deferred until first call.
        """

        filename = self.create_path(session_date, test_name)
        super(S_FileHandler, self).__init__(filename, mode, encoding, delay)
        self.setFormatter(Formatter(text_format, date_format))

    def create_path(self, session_date: str, test_name: str) -> Path:
        """
        Creates both test session and log file paths.

        Args:
            session_date (str): Start date of whole test session.
            test_name (str): From test module, starts with "test_*".

        Returns:
            Path: Full path of log file.
        """
        folder_path = os.path.join('logs', f'test_session_{session_date}')
        file_path = os.path.join(folder_path, f'{test_name}.log')

        if not os.path.exists(folder_path):
            os.makedirs(folder_path)

        if not os.path.exists(file_path):
            with open(file_path, 'w+'):
                pass

        return file_path


class S_StreamHandler(StreamHandler):
    """
    Class that overrides standard Logging StreamHanlder.
    """

    def __init__(self) -> None:
        """
        Initialization of StreamHandler instance.
        """
        StreamHandler.__init__(self)
        formatter = Formatter(text_format, date_format)
        self.setFormatter(formatter)


class S_Logger(Logger):
    """
    Class that overrides standard Logging library.
    """
    INFO = 20
    ERROR = 40

    def __init__(self,
                 test_name: str,
                 start_time: str,
                 log_name: str = 'main',
                 *args, **kwargs) -> None:
        """
        Initialization of Logger instance.

        Args:
            test_name (str): From test module, starts with "test_*".
            start_time (str): Test Session start time.
            log_name (str, optional): Logger name. Defaults to 'main'.
        """
        Logger.__init__(self, log_name, *args, **kwargs)
        self.logger = logging.getLogger(log_name)
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(S_StreamHandler())
        self.logger.addHandler(S_FileHandler(test_name, start_time))

        self.test_name = test_name
        self.setLevel(logging.DEBUG)
        self.addHandler(S_StreamHandler())
        self.addHandler(S_FileHandler(test_name, start_time))

    def start(self, *args) -> None:
        """
        Test start logging method.
        """
        if self.isEnabledFor(S_Logger.INFO):
            self._log(S_Logger.INFO,
                      f'THE EXECUTION OF {self.test_name} HAS STARTED.',
                      args)

    def end(self, start: time.time, status: Status, *args) -> float:
        """
        Test end logging method. Computes duration
        of test execution.

        Args:
            start (time.time): Test start time.

        Returns:
            float: Test execution duration.
        """
        end = time.perf_counter()
        duration = round(end-start, 2)

        if self.isEnabledFor(S_Logger.INFO):
            self._log(S_Logger.INFO,
                      (f'THE EXECUTION OF {self.test_name}' +
                       f'HAS ENDED WITH RESULT: {status.name}'),
                      args)

            self._log(S_Logger.INFO,
                      f'FINISHED in {duration} second(s)',
                      args)

        return duration

    def exception(self, traceback: str, *args) -> None:
        """
        Test exception logging method.

        Args:
            traceback (str): Full exception traceback in string from.
        """
        if self.isEnabledFor(S_Logger.ERROR):
            self._log(S_Logger.ERROR,
                      f'AN EXCEPTION OCCURRED: {traceback}',
                      args)

    @classmethod
    def get_logger(cls) -> logging.Logger:
        """
        Getter for Logger instance.

        Returns:
            logging.Logger: Logger instance.
        """
        return logging.getLogger('main')



import os
import ast

from typing import Tuple
from source.exceptions import (NoTestsFoundError,
                               TestOutOfClassError)


def find_folder_path() -> str:
    """
    Searches for main test folder path.

    Returns:
        str: path to test folder
    """
    for root, dirs, _ in os.walk(".", topdown=True):
        exclude = set(['.git', '.venv', '.vscode', '__pycache__'])
        dirs[:] = [d for d in dirs if d not in exclude]

        for dirname in dirs:
            if dirname in ('tests', 'test'):
                return os.path.join(root, dirname)

        raise FileNotFoundError('Could not find test(s) folder.')


def get_test_modules() -> list[str]:
    """
    Searches for test files recursively in given directory.
    Test file must start with "test_" to be recognized.

    Returns:
        list: list of test module names
    """
    test_files = []

    for root, _, files in os.walk(find_folder_path()):
        for file_name in files:

            # Ignore cache files
            if file_name.endswith('.pyc'):
                continue

            # Look only for those that starts with:
            # FIXME Fix path filtering for UNIX.
            if file_name.startswith('test_'):
                file_path = os.path.join(root[2:], file_name)
                test_files.append(file_path)

    if test_files:
        return test_files
    raise NoTestsFoundError


def show_info(function_node: ast.FunctionDef) -> None:
    """
    Prints out information about given function
    and its parameters.

    Args:
        function_node (ast.FunctionDef): Function definition.
    """
    if not isinstance(function_node, ast.FunctionDef):
        raise TypeError(f'{function_node} must be type of ast.FunctionDef')

    print(f"Function name: {function_node.name}")
    if function_node.args.args:
        print("Args:")

        for arg in function_node.args.args:
            print(f"\tParameter name: {arg.arg}")


def read_from_module(module: str) -> Tuple[ast.FunctionDef, ast.ClassDef]:
    """
    Opens module and gets all functions and classes.

    Returns:
        Tuple[ast.FunctionDef, ast.ClassDef]: Functions and classes.
    """
    with open(module, encoding='utf-8') as file:
        node = ast.parse(file.read())

    functions = [n for n in node.body if isinstance(n, ast.FunctionDef)]
    classes = [n for n in node.body if isinstance(n, ast.ClassDef)]

    return functions, classes


def create_tree() -> list[dict]:
    """
    Searches for test functions in given test modules.

    Returns:
        list[str]: List of module names along
                    with their test functions.
    """
    test_tree = []

    for module in get_test_modules():

        functions, classes = read_from_module(module)

        module = module.replace(os.sep, '.').replace('.py', '')
        dict = {}
        dict[module] = []

        # If test function detected raise exception.
        for function in functions:
            if function.name.startswith('test_'):
                raise TestOutOfClassError(function.name)

        # Search for tests in test classes.
        for class_ in classes:
            dict[module].append({class_.name: []})
            methods = [n for n in class_.body if isinstance(n, ast.FunctionDef)]
            # TODO Create way of adding info about decorator to the dict.

            # Add methods to nested dictionary.
            for method in methods:
                if method.name.startswith('test_'):
                    if method.decorator_list:
                        pass
                        # print(method.decorator_list[0].id)
                    dict[module][-1][class_.name].append(method.name)

        test_tree.append(dict)

    return test_tree



class Error(Exception):
    """
    Base class for other exceptions.
    """
    def __init__(self) -> None:
        self.message = 'Base Custom Error Class'

    def __str__(self) -> str:
        return self.message


class EmptyFileError(Error):
    """
    Raised when the opened file is empty.
    """
    def __init__(self, file_name) -> None:
        self.message = f"'{file_name}' file is empty."


class NoTestsFoundError(Error):
    """
    Raised when no tests were found.
    """
    def __init__(self) -> None:
        self.message = "No tests found in the test folder."


class TestOutOfClassError(Error):
    """
    Raised when test function was discovered out of Test Class.
    """
    def __init__(self, test_name) -> None:
        self.message = f"'{test_name}' test function out of Test Class."



import inspect
import functools

from typing import Any, Callable


class Template(object):
    """
    Template for specific decorator creation.
    """
    def __init__(self, object) -> None:

        self.object = object
        functools.update_wrapper(self, object)

    def __call__(self, *args: Any, **kwargs: Any) -> Callable:
        """
        When decorator is being called check if is used
        on whole class or single method. If class then
        decorate all methods, otherwise function.

        Returns:
            Callable: Decorated function or method.
        """
        if inspect.isclass(self.object):
            for name, method in inspect.getmembers(self.object):
                if not inspect.isfunction(method) \
                   or inspect.isbuiltin(method):
                    continue

                if not inspect.ismethod(method) and \
                   not name.startswith('test_'):
                    continue

                setattr(self.object, name, self.decorate_method(method))
            return self.object(*args, **kwargs)

        else:
            self.object(*args, **kwargs)

    def decorate_method(self, original_func: Callable) -> Callable:
        """
        Helper function for decorating particular class method.

        Args:
            original_func (Callable): _description_

        Returns:
            Callable: Decorated method.
        """
        @functools.wraps(original_func)
        def decorator(*args, **kwargs):
            return original_func(self, *args, **kwargs)
        return decorator


SMOKE_TEST = Template
REGRESSION_TEST = Template



import importlib

from ast import Module
from typing import Tuple
from multiprocess import Semaphore
from source.testcase import Results
from source.process import S_Process
from source.utilities import get_time
from source.searcher import create_tree


class Runner(object):
    """
    Class responsible for creating processes in which
    each test is executed separately and independently.
    """

    def __init__(self) -> None:
        """
        Initialization of processes list
        and generating tests' hierarchy.
        """
        self.processes: list = []
        self.test_tree: list[dict] = create_tree()
        self.test_results = Results()

    def collect_tests(self) -> None:
        """
        Method for future test selection.
        """
        pass

    def importer(self, module: dict) -> Tuple[Module, str, str]:
        """
        Imports given module name.

        Args:
            module (dict): Module in form of dictionary
            with name and its class members.

        Raises:
            ModuleNotFoundError: When module was not found.

        Returns:
            Tuple[Module, str, str]: Imported module,
            its name and members.
        """
        mod_name = next(iter(module.keys()))
        mod_members = next(iter(module.values()))

        try:
            mod = importlib.import_module(mod_name)
        except ModuleNotFoundError as Error:
            raise Error

        return mod, mod_name, mod_members

    def run_tests(self) -> None:
        """
        Creates independent processes and appends
        them into the pool of processes.
        """
        # Create a pool of processes by defined number
        # ? NO_SYSTEM_CPU = multiprocess.cpu_count()
        concurrency = 2
        semaphore = Semaphore(concurrency)
        start_time = get_time()

        # For each module in test hierarchy.
        for module in self.test_tree:
            module, mod_name, mod_members = self.importer(module)

            # For each TestSuite in module.
            for test_class in mod_members:
                _class = getattr(module, next(iter(test_class.keys())))

                # For each test in given TestSuite.
                for test_name in test_class[next(iter(test_class.keys()))]:
                    _test_instance = _class(test_name)

                    semaphore.acquire()

                    process = S_Process(
                            target=getattr(_class, test_name),
                            args=(_test_instance,),
                            test_name=test_name,
                            start_time=start_time,
                            semaphore=semaphore
                    )
                    process.start()
                    self.processes.append((process, _test_instance))

        self.get_results()

    def get_results(self) -> None:
        """
        Wait untill all processes are finished
        and get the test session results.
        """
        for process, instance in self.processes:
            process.join()

            instance.status = process.result[1]
            instance.duration = process.result[2]
            instance.reruns = process.result[3]

            self.test_results.add(instance)

    def show_results(self) -> None:
        """
        Print out results.
        """
        for test in self.test_results.tests:
            print(f'{test.name: <15} \
                    {test.status: <15} \
                    {test.duration: <5} \
                    {test.reruns}')


if __name__ == "__main__":
    runner = Runner()
    # runner.collect_tests()
    runner.run_tests()
    runner.show_results()



import time

from traceback import format_exc
from source.logger import S_Logger
from source.testcase import Status
from multiprocess import (Semaphore,
                          Process,
                          Pipe)


class S_Process(Process):
    """
    Process class overrides standard multiprocess.Process
    This class is responsible for managing and running
    everything what is related with running tests in Process.
    """

    def __init__(self,
                 test_name: str,
                 start_time: str,
                 semaphore: Semaphore,
                 *args,
                 **kwargs) -> None:
        """
        Initialization of Process instance.
        Creates multiprocess.Pipe which allows to
        communicate with parent and child processes.

        Args:
            test_name (str): From test module, starts with "test_*".
            start_time (str): Test Session start time.
            semaphore (Semaphore): Manages an internal counter
                                   of available processes.
        """
        Process.__init__(self, *args, **kwargs)
        self._parent_conn, self._child_conn = Pipe()
        self._duration: float = 0
        self.test_name = test_name
        self.semaphore = semaphore
        self.start_time = start_time
        self.status = Status.PASS

    def run(self) -> None:
        """
        Starts process along with logger. Sends
        response back to parent process through Pipe.
        """
        log = S_Logger(self.test_name, self.start_time)
        NO_RUN = 0
        MAX_RUNS = 3 # TODO Make that a config value

        while NO_RUN < MAX_RUNS:
            start = time.perf_counter()
            NO_RUN += 1

            try:
                log.start()
                Process.run(self)

            except Exception as e:
                log.exception(format_exc())
                if isinstance(e, AssertionError):
                    self.status = Status.FAIL
                else:
                    self.status = Status.ERROR
            else:
                break

            finally:
                duration = log.end(start, self.status)
                self._child_conn.send((self.test_name,
                                       self.status,
                                       duration,
                                       NO_RUN))
                self.semaphore.release()

    @property
    def result(self) -> None:
        """
        Result getter for parent process.
        """
        if self._parent_conn.poll():
            self._result = self._parent_conn.recv()
        return self._result



from enum import Enum, auto
from typing import Optional, Tuple


class MetaTestCase(type):
    """
    Metaclass for changing behaviour of
    TestCase class in order to add before/after
    functionalities during test execution.
    """

    def __new__(cls, name: str, bases: Tuple, attrs: dict):
        """
        When creating a new object of that class
        we create a new TestCase class with
        modified properties.

        Args:
            name (str): Class name that invokation came from
            bases (Tuple): Base classes that were inherited
            attrs (dict): Attributes of invoking class
        """
        def replaced_func(fn):
            def new_test(*args, **kwargs):
                args[0].before()
                result = fn(*args, **kwargs)
                args[0].after()
                return result

            return new_test

        # If method is found and its name starts with test, replace it
        for i in attrs:
            if callable(attrs[i]) and attrs[i].__name__.startswith('test'):
                attrs[i] = replaced_func(attrs[i])

        return (super(MetaTestCase, cls).__new__(cls, name, bases, attrs))


class Status(Enum):
    """
    The status of a given test or test session.
    """

    PASS = auto()
    FAIL = auto()
    ERROR = auto()


class TestCase(metaclass=MetaTestCase):
    """
    An individual Test Case.
    """

    def __init__(self, name) -> None:
        self.name = name
        self._duration: float = 0
        self.message: Optional[str]
        self._status: Status = Status.PASS
        self._reruns: int = 0

    def before(self) -> None:
        """
        Overridable, execute before test part.
        """
        pass

    def after(self) -> None:
        """
        Overridable, execute after test part.
        """
        pass

    def run(self) -> None:
        """
        Execute test part.
        """
        pass

    def _update(self, status: Status, message: str) -> None:
        """
        Internal, update status property.
        """
        self._status = status
        self.message = message

    def fail(self, message: str | None) -> None:
        """
        Indicate this test failed.
        """
        self._update(Status.FAIL, message)

    def error(self, message: str | None) -> None:
        """
        Indicate this test has encountered an error.
        """
        self._update(Status.ERROR, message)

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status: Status):
        self._status = status

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, duration: float):
        self._duration = duration

    @property
    def reruns(self):
        return self._reruns

    @reruns.setter
    def reruns(self, reruns: float):
        self._reruns = reruns


class Results(object):
    """
    Overall results of a test run.
    """

    def __init__(self,
                 status: Status = Status.PASS,
                 message: str | None = None
                 ) -> None:
        self.status = status
        self.message = message
        self.tests = []

    def add(self, test: TestCase) -> None:
        """
        Add a Test to the list of tests.
        """
        if test.status is Status.FAIL:
            self.fail()

        self.tests.append(test)

    def fail(self) -> None:
        """
        Indicate the test run had at least one failure.
        """
        self.status = Status.FAIL

    def error(self, message: str | None = None) -> None:
        """
        Indicate the test run fatally errored.
        """
        self.status = Status.ERROR
        self.message = message
